from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Tuple

import requests
import tomli as toml
import typer
from pysh import sh, which

from . import update
from .config import CONFIG, get_gitlab_api_url
from .utils.common import cd_with_log, clean_python_version, parse_python_version, sh_with_log

PYENV_INSTALLED = which("pyenv")


@CONFIG.requires("git_url")
def setup_services(services: List[str], skip_existing: bool) -> None:
    if not "git_url" in CONFIG:
        raise typer.BadParameter("You must set the git_url in the config file before you can use this command")
    if not PYENV_INSTALLED:
        typer.echo("Failed to locate pyenv. Will use the system python version(s) instead", err=True)
    installable_pyenv_versions = [
        clean_python_version(" \t~^*") for v in sh("pyenv install --list", capture=True).stdout.split("\n") if v.strip()
    ]

    repositories_to_clone = get_repositories_to_clone(services, skip_existing)

    for name, link in repositories_to_clone:
        setup_service(name, link, installable_pyenv_versions)


def setup_service(repo_name: str, git_link: str, installable_pyenv_versions: List[str]):
    typer.echo(f"Setting up {repo_name}", err=True)
    success = clone_repo(git_link)

    if success:
        with cd_with_log(repo_name):
            if Path("pyproject.toml").exists():
                python_version = get_python_version(Path("pyproject.toml"))
                if python_version:
                    if PYENV_INSTALLED:
                        setup_pyenv_locally(python_version, installable_pyenv_versions)
                    else:
                        sh_with_log(f"poetry env use {python_version}")
                sh_with_log("poetry install --all-extras")
            update.env([Path()])
            update.ports([Path()])
    else:
        raise FileNotFoundError(f"Failed to enter the directory '{repo_name}'")


def clone_repo(git_link: str) -> bool:
    return bool(sh_with_log(f"git clone {git_link}"))


def get_python_version(pyproject_path: Path) -> Optional[str]:
    pyproject = toml.loads(pyproject_path.read_text())
    raw_python_version: str = pyproject["tool"]["poetry"]["dependencies"]["python"]  # type: ignore
    version_info = parse_python_version(raw_python_version)
    if version_info:
        return f"{version_info[0]}.{version_info[1]}"
    else:
        typer.echo(
            "Failed to read python version from pyproject.toml. It's either invalid or too complex for me.", err=True
        )


def setup_pyenv_locally(python_version: str, installable_pyenv_versions: List[str]):
    raw_installed_pyenv_venvs = sh("pyenv versions", capture=True).stdout.split("\n")
    installed_pyenv_venvs = [clean_python_version(v) for v in raw_installed_pyenv_venvs]
    installable_python_version = get_usable_pyenv_version(
        python_version,
        installed_pyenv_venvs,
    ) or get_usable_pyenv_version(
        python_version,
        installable_pyenv_versions,
    )

    if installable_python_version is not None:
        if " " in installable_python_version:
            installable_python_version = installable_python_version.split(" ")[0]
        sh_with_log(f"pyenv local {installable_python_version}")
        sh_with_log(f"poetry env use {installable_python_version}")


def get_usable_pyenv_version(current: str, available: Sequence[str], install: bool = False) -> Optional[str]:
    for version in reversed(available):
        version = version.strip("* \t\n")
        if version.startswith(current):
            if install:
                sh(f"pyenv install {version}")
            return version


def get_repositories_to_clone(repo_names: List[str], skip_existing: bool) -> List[Tuple[str, str]]:
    """Expands gitlab repo names to include all repos if the name is a group or a namespace"""
    expanded_repo_names = []

    for name in repo_names:
        # Remove all whitespace in case the user accidentally added some
        name = "".join(name.split())

        if name.count("/") == 2:
            expanded_repo_names.append((name.split("/")[-1], f'{CONFIG["git_url"]}:{name}.git'))
        else:
            with requests.Session() as session:
                session.headers = {"PRIVATE-TOKEN": CONFIG.get_api_token()}
                project_jsons = _paginated_get(f"{get_gitlab_api_url()}/projects", session)

                projects = [
                    (p["path"], p["ssh_url_to_repo"])
                    for p in project_jsons
                    if p["path_with_namespace"].startswith(name)
                ]
                if not projects:
                    raise ValueError(
                        f"Failed to find any projects that start with '{name}'. Maybe you need to add a group/namespace or to fix a typo?"
                    )
                non_repeating_projects = []
                for project in projects:
                    if Path(project[0]).exists():
                        yes = skip_existing or typer.confirm(
                            f"Found project {project[0]} but a folder with the same name already exists. You should use `stb update` for it instead. Would you like to skip it?",
                        )
                        if not yes:
                            raise typer.Exit(1)
                    else:
                        non_repeating_projects.append(project)
                expanded_repo_names.extend(non_repeating_projects)

    return expanded_repo_names


def _paginated_get(url: str, session: requests.Session) -> List[Dict[str, Any]]:
    """Gets a paginated response from the gitlab api"""
    response = session.get(url, params={"per_page": 20})
    response.raise_for_status()

    total_pages = int(response.headers["X-Total-Pages"])

    projects = response.json()

    for _ in range(1, total_pages):
        response = session.get(response.links["next"]["url"])
        response.raise_for_status()
        projects.extend(response.json())

    return projects



from contextlib import suppress
from pathlib import Path
from typing import List

import rich
import typer
from pysh import sh

from .db import Choices
from .db import run_on_single_service as stb_db
from .utils.common import (
    ENV_VARS,
    SERVICE_PATHS_ARG,
    add_default_service_path,
    cd_with_log,
    gather_services,
    save_dotenv_file,
    sh_with_log,
)


def old_reset_databases_flag_deprecation_callback(value: bool) -> None:
    if value:
        rich.print(
            "[bold red]The flag '-d' for 'stb update package' is deprecated. It ran 'stb db reset -fp' as well but now this is the default behavior. If you want to disable it, use --no-reset-databases.[/bold red]\n"
        )


app = typer.Typer(
    name="update",
    help="Updates your services' configurations, dependencies, submodules, and ports without losing any part of your environment or configurations",
)

OLD_RESET_DATABASES_ARG = typer.Option(
    False,
    "-d",
    "--reset-databases",
    help="This option is deprecated. It ran 'stb db reset -fp' as well but now this is the default behavior. If you want to disable it, use --no-reset-databases.",
    callback=old_reset_databases_flag_deprecation_callback,
)
NO_RESET_DATABASES_ARG = typer.Option(
    False,
    "-D",
    "--no-reset-databases",
    help="Do not run 'stb db reset -fp' after updating the services",
)


@app.command()
@add_default_service_path
def env(service_paths: List[Path] = SERVICE_PATHS_ARG):
    """Update .env files with new/modified fields from .env.example"""
    for service in gather_services(service_paths).values():
        for field, example_value in service.dotenv_example.items():
            replacing_value = example_value or ENV_VARS.get(field)
            if field in service.dotenv:
                pass
            else:
                service.dotenv[field] = str(replacing_value) if replacing_value else ""
        save_dotenv_file(service)
        typer.echo(f"Updated {service.dotenv_path}")


@app.command()
@add_default_service_path
def ports(service_paths: List[Path] = SERVICE_PATHS_ARG) -> None:
    """I update service ports to allow you to quickly set up a set of microservices locally and use all others from dev"""
    services = gather_services(service_paths)
    service_to_port_mapper = {service.dir.name: port for port, service in enumerate(services.values(), start=8000)}
    microservice_fields = {convert_microservice_name_to_env_field(m): n for m, n in service_to_port_mapper.items()}
    for service_name, service in services.items():
        helm_env_defaults = (service.yaml_config or {}).get("common", {}).get("envs", {})
        service.dotenv["SERVICE_PORT"] = str(service_to_port_mapper[service_name])
        for field in [f for f in service.dotenv if f.endswith("_URL")]:
            if field in microservice_fields:
                service.dotenv[field] = f"http://localhost:{microservice_fields[field]}"
            else:
                helm_defaults = helm_env_defaults.get(field.upper())
                if helm_defaults is not None and "review" in helm_defaults:
                    service.dotenv[field] = helm_defaults["review"]

        save_dotenv_file(service)
        typer.echo(f"Updated {service.dotenv_path}")


@app.command()
@add_default_service_path
def package(
    service_paths: List[Path] = SERVICE_PATHS_ARG,
    install: bool = typer.Option(
        True,
        help="Run poetry install",
    ),
    all_extras: bool = typer.Option(
        True,
        help="If --install-dependencies is true, install all extras as well",
    ),
    update_dependencies: bool = typer.Option(
        False,
        "-u",
        "--update",
        help="Run poetry update instead of poetry install",
    ),
    pull_changes: bool = typer.Option(
        False,
        "-p",
        "--pull",
        help="Pull changes from remote",
    ),
    update_ports: bool = typer.Option(
        False,
        "-P",
        "--ports",
        help="Update service ports to allow ysou to quickly set up a set of microservices locally and use all others from dev",
    ),
    update_env: bool = typer.Option(
        False,
        "-e",
        "--env",
        help="Update .env files with new/modified fields from .env.example",
    ),
    checkout_to_master: bool = typer.Option(
        False,
        "-c",
        "--checkout",
        help="Stash the current changes and checkout to master",
    ),
    old_reset_databases: bool = OLD_RESET_DATABASES_ARG,
    no_reset_databases: bool = NO_RESET_DATABASES_ARG,
):
    """Install the dependencies from poetry.lock file, update submodules, optionally update dependencies, and optionally reset databases"""
    branches_where_stashes_happened = []

    for service in gather_services(service_paths).values():
        with cd_with_log(service.dir):
            if checkout_to_master:
                res = sh("git diff", capture=True)
                if res.stdout:
                    sh_with_log("git stash")
                    res = sh("git branch --show-current", capture=True)
                    if res.returncode == 0:
                        branches_where_stashes_happened.append(service.dir.name + "/" + res.stdout.strip())
                sh_with_log("git checkout master")
            if pull_changes:
                sh_with_log("git pull")
            if update_dependencies:
                sh_with_log("poetry update")
            elif install:
                all_extras_arg = "--all-extras" if all_extras else ""
                sh_with_log(f"poetry install {all_extras_arg}")
            if update_env:
                env([service.dir])
            if update_ports:
                ports([service.dir])
            if not no_reset_databases:
                print(f"Resetting databases for {service.dir.name}...")
                with suppress(LookupError):
                    stb_db(service.dir, Choices.drop, force_drop=True)
                    stb_db(service.dir, Choices.create, parallel_migrations=True)

    if branches_where_stashes_happened:
        typer.echo(
            f"------------\nStashed changes in the following branches: {', '.join(branches_where_stashes_happened)}"
        )


def convert_microservice_name_to_env_field(name: str) -> str:
    return name.replace("-", "_").strip().upper() + "_URL"


if __name__ == "__main__":
    app()



from typing import Set

import typer

from stb.utils.common import cd_with_log, sh_with_log

app = typer.Typer(
    name="run",
    help="Runs the select services together at the same time",
)


def run_services(services: Set[str]) -> None:
    typer.echo("Checking out services...", err=True)

    for service in services:
        with cd_with_log(service):
            # TODO: Sounds like this needs to reuse `stb update package`
            sh_with_log("git stash")
            sh_with_log("git checkout master")
            sh_with_log("git pull")
            sh_with_log("stb db reset")
            sh_with_log("poetry install --all-extras")

    concurrently_query = " ".join([f'"cd {s} && (make run || poetry run python3 run.py)"' for s in services])

    sh_with_log("concurrently " + concurrently_query)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['stb', 'stb.utils']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'keyring>=23.13.1,<24.0.0',
 'platformdirs>=2.6.2,<3.0.0',
 'pysh>=3.1.0,<4.0.0',
 'python-dotenv>=0.21.0,<0.22.0',
 'python-gitlab>=3.13.0,<4.0.0',
 'requests>=2.28.2,<3.0.0',
 'tomli>=2.0.1,<3.0.0',
 'tomlkit>=0.11.6,<0.12.0',
 'typer[all]>=0.6.1,<0.7.0',
 'typing-extensions>=4.3.0,<5.0.0']

entry_points = \
{'console_scripts': ['stb = stb:app']}

setup_kwargs = {
    'name': 'stb-mnt',
    'version': '4.5.0',
    'description': 'A universal tool for local microservice management.',
    'long_description': '# stb\n\nA universal tool for local microservice management\n\n## Installation\n\n```bash\npipx install stb-mnt\n```\n\n## Usage\n\n### Setup\n\n* To download and setup my_company/backend/service1 microservice as a subdirectory to the current working directory, use:\n\n```bash\nstb setup my_company/backend/service1\n```\n\n* To download and setup my_company/backend/service1 and my_company/backend/service2 as subdirectories to current working directory, use:\n\n```bash\nstb setup my_company/backend/service1 my_company/backend/service2\n```\n\n* To setup all backend services, use:\n\n```bash\nstb setup my_company/backend\n```\n\nNote that if you want to clone repositories, you must first set a `git_url` using `stb config set git_url` command\n\n### Update\n\n* To update .env file in accordance with .env.example in a microservice:\n\n```bash\nstb update env\n```\n\n* To synchronize service ports between all installed microservices (you can specify which ones will run locally with the `--local` option):\n\n```bash\nstb update ports\n```\n\n* To update poetry.lock file, install dependencies, stash current changes, checkout to master, pull from remote, and recreate databases:\n\n```bash\nstb update package -piucd\n```\n\nor  \n\n```bash\nstb update package --pull --update --checkout --reset-databases\n```\n\n### DB\n\n* To upgrade migrations in a microservice:\n\n```bash\nstb db upgrade\n```\n\n* To create databases and upgrade its migrations in a microservice:\n\n```bash\nstb db create\n```\n\n* To drop databases in a microservice:\n\n```bash\nstb db drop\n```\n\n* To drop and recreate databases, and upgrade migrations in a microservice:\n\n```bash\nstb db reset\n```\n\n* To upgrade migrations in parallel for faster upgrades (useful for large monoliths with multiple databases), you can use the -p (--parallel) option:\n\n```bash\nstb db create -p\n```\n  \n```bash\nstb db reset -p\n```\n  \n* To force dropping of databases in case another program is using them at the same time, you can use the -f (--force) option:\n\n```bash\nstb db drop -f\n```\n  \n```bash\nstb db reset -f\n```\n  \n### Use\n\n`stb use` allows you to take a company private package and install either a cloud version or a local version of it. STB will preserve all extras, automatically set package source, and will gracefully handle any issues that might happen while updating.\n\n* To install a local version of `my_package` that is located at `../my_package`:\n\n```bash\nstb use ../my_package\n```\n\n* To install a local version of `my_package` that is located at `../my_package` in editable mode:\n\n```bash\nstb use ../my_package --editable\n```\n\n* To install a cloud version of `my_package` with tag `8.3.1`:\n\n```bash\nstb use "my_package==8.3.1"\n```\n\n* To install a cloud version of my_package with tag `8.3.1`, my_other_package with any tag higher than `1.2.3`, and my_third_package with any tag more than or equal to `4.5.6` and less than `5.0.0`:\n\n```bash\nstb use "my_package==8.3.1" "my_other_package>1.2.3" "my_third_package^4.5.6"\n```\n\n### Run\n\n* To update and run the select services concurrently:\n\n```bash\nstb run service1 service2\n```\n\n### Config\n\n* To set a git url for cloning:\n\n```bash\nstb config set git_url git@gitlab.my_company.com\n```\n\n### Graph\n\n* To get a dependency graph of your microservices:\n\n```bash\nstb graph json my_company/backend/ my_company/infrastructure/\n```\n\n* To get a dependency graph of your microservices as an svg image (requires graphviz):\n\n```bash\nstb graph graphviz my_company/backend/ my_company/infrastructure/\n```\n\n### How directories are selected for update/db\n\nFor every update, you can specify:\n\n1) A microservice directory, which will cause stb to update only that microservice\n2) Several microservice directories, which will cause stb to update these microservices and integrate them together (for example, `update ports` assigns ports to local microservices and updates their links in other microservices to match the assigned ports)\n3) A directory with multiple microservice subdirectories inside it, which is equivalent to (2) with the list of subdirectories as arguments\n4) Nothing, which will choose the current working directory as the first argument and will be equivalent to (1) or (3)\n',
    'author': 'Stanislav Zmiev',
    'author_email': 'szmiev2000@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
